#!/usr/bin/env python3

import os
import shutil
import sys

BASE = 'https://www.monetdb.org/downloads/Windows/Latest/'

if len(sys.argv) != 3:
    exit(f"Usage: {sys.argv[0]} <STAGING> <DEST>")
staging_dir = sys.argv[1]
dest_dir = sys.argv[2]

TREE = """
bin/
bin/bat.dll
bin/bat.pdb
bin/bz2.dll
bin/charset-1.dll
bin/geos.dll
bin/geos_c.dll
bin/getopt.dll
bin/iconv-2.dll
bin/libxml2.dll
bin/lz4.dll
bin/lzma.dll
bin/mapi.dll
bin/mapi.pdb
bin/mclient.exe
bin/mclient.pdb
bin/monetdb5.dll
bin/monetdb5.pdb
bin/monetdbe.dll
bin/monetdbsql.dll
bin/mserver5.exe
bin/mserver5.pdb
bin/msqldump.exe
bin/msqldump.pdb
bin/pcre.dll
bin/stream.dll
bin/stream.pdb
bin/zlib1.dll
etc/
etc/.monetdb
include/
include/monetdb/
include/monetdb/copybinary.h
include/monetdb/exception_buffer.h
include/monetdb/gdk.h
include/monetdb/gdk_atoms.h
include/monetdb/gdk_bbp.h
include/monetdb/gdk_calc.h
include/monetdb/gdk_cand.h
include/monetdb/gdk_delta.h
include/monetdb/gdk_hash.h
include/monetdb/gdk_posix.h
include/monetdb/gdk_strimps.h
include/monetdb/gdk_system.h
include/monetdb/gdk_time.h
include/monetdb/gdk_tracer.h
include/monetdb/gdk_utils.h
include/monetdb/mal.h
include/monetdb/mal_authorize.h
include/monetdb/mal_client.h
include/monetdb/mal_errors.h
include/monetdb/mal_exception.h
include/monetdb/mal_function.h
include/monetdb/mal_import.h
include/monetdb/mal_instruction.h
include/monetdb/mal_linker.h
include/monetdb/mal_listing.h
include/monetdb/mal_module.h
include/monetdb/mal_namespace.h
include/monetdb/mal_prelude.h
include/monetdb/mal_resolve.h
include/monetdb/mal_stack.h
include/monetdb/mal_type.h
include/monetdb/mapi.h
include/monetdb/matomic.h
include/monetdb/mel.h
include/monetdb/monet_getopt.h
include/monetdb/monet_options.h
include/monetdb/monetdb_config.h
include/monetdb/monetdbe.h
include/monetdb/mstring.h
include/monetdb/opt_backend.h
include/monetdb/rel_basetable.h
include/monetdb/rel_distribute.h
include/monetdb/rel_dump.h
include/monetdb/rel_exp.h
include/monetdb/rel_optimizer.h
include/monetdb/rel_partition.h
include/monetdb/rel_prop.h
include/monetdb/rel_rel.h
include/monetdb/rel_semantic.h
include/monetdb/sql_atom.h
include/monetdb/sql_backend.h
include/monetdb/sql_catalog.h
include/monetdb/sql_hash.h
include/monetdb/sql_import.h
include/monetdb/sql_keyword.h
include/monetdb/sql_list.h
include/monetdb/sql_mem.h
include/monetdb/sql_mvc.h
include/monetdb/sql_parser.h
include/monetdb/sql_privileges.h
include/monetdb/sql_qc.h
include/monetdb/sql_query.h
include/monetdb/sql_relation.h
include/monetdb/sql_scan.h
include/monetdb/sql_semantic.h
include/monetdb/sql_stack.h
include/monetdb/sql_storage.h
include/monetdb/sql_string.h
include/monetdb/sql_symbol.h
include/monetdb/sql_tokens.h
include/monetdb/sql_types.h
include/monetdb/store_sequence.h
include/monetdb/stream.h
include/monetdb/stream_socket.h
lib/
lib/bat.lib
lib/bz2.lib
lib/charset.lib
lib/getopt.lib
lib/iconv.lib
lib/libxml2.lib
lib/lz4.lib
lib/lzma.lib
lib/mapi.lib
lib/monetdb5.lib
lib/monetdb5/
lib/monetdb5/_generator.dll
lib/monetdb5/_generator.pdb
lib/monetdb5/_geom.dll
lib/monetdb5/_geom.pdb
lib/monetdb5/_pyapi3.dll
lib/monetdb5/_pyapi3.pdb
lib/monetdbe.lib
lib/monetdbsql.lib
lib/pcre.lib
lib/stream.lib
lib/zlib.lib
license.rtf
M5server.bat
mclient.bat
msqldump.bat
MSQLserver.bat
pyapi_locatepython3.bat
share/
share/doc/
share/doc/MonetDB-SQL/
share/doc/MonetDB-SQL/dump-restore.html
share/doc/MonetDB-SQL/dump-restore.txt
share/doc/MonetDB-SQL/website.html
System64/
System64/concrt140.dll
System64/msvcp140.dll
System64/msvcp140_1.dll
System64/msvcp140_2.dll
System64/msvcp140_atomic_wait.dll
System64/msvcp140_codecvt_ids.dll
System64/vccorlib140.dll
System64/vcruntime140.dll
System64/vcruntime140_1.dll
"""


failures = 0
for line in TREE.strip().splitlines():
    parts = line.split('/')
    if not parts[-1]:
        continue
    src0 = parts[-1]
    src0 = src0.replace('-', '_')
    if src0.startswith('.'):
        src0 = '_' + src0
    if '140' in src0 and '.dll' in src0:
        src0 += '.DFEFC2FE_EEE6_424C_841B_D4E66F0C84A3'
    src = os.path.join(staging_dir, src0)
    tgt_dir = os.path.join(dest_dir, *parts[:-1])
    tgt = os.path.join(dest_dir, *parts)
    if not os.path.isdir(tgt_dir):
        print(f"Creating dir {tgt_dir}", flush=True)
        os.makedirs(tgt_dir)
    print(f"Copying [{src0}] {src} to {tgt}", flush=True)
    try:
        shutil.copyfile(src, tgt)
    except Exception as e:
        print(f"  !! FAILED: {e}", flush=True)
        failures += 1

if failures:
    exit(f"Encountered {failures} failures")



# coding: utf-8

# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0.  If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
# Copyright 1997 - July 2008 CWI, August 2008 - 2016 MonetDB B.V.

import unittest
import pymonetdb
from tests.util import test_args


class TestUnicode(unittest.TestCase):
    table_prefix = 'dbapi20test_'  # If you need to specify a prefix for tables

    ddl1 = 'create table %sbooze (name varchar(60))' % table_prefix
    ddl2 = 'create table %sbarflys (name varchar(60))' % table_prefix
    xddl1 = 'drop table %sbooze' % table_prefix
    xddl2 = 'drop table %sbarflys' % table_prefix

    lowerfunc = 'lower'  # Name of stored procedure to convert string->lowercase

    # Some drivers may need to override these helpers, for example adding
    # a 'commit' after the execute.
    def executeDDL1(self, cursor):
        cursor.execute(self.ddl1)

    def executeDDL2(self, cursor):
        cursor.execute(self.ddl2)

    def setUp(self):
        pass

    def tearDown(self):
        con = self._connect()
        try:
            cur = con.cursor()
            for ddl in (self.xddl1, self.xddl2):
                try:
                    cur.execute(ddl)
                    con.commit()
                except pymonetdb.Error:
                    # Assume table didn't exist. Other tests will check if
                    # execute is busted.
                    pass
        finally:
            con.close()

    def _connect(self):
        try:
            return pymonetdb.connect(**test_args)
        except AttributeError:
            self.fail("No connect method found in pymonetdb module")

    def test_unicode_string(self):
        con = self._connect()
        cursor = con.cursor()
        self.executeDDL1(cursor)
        x = u"ô  ’a élé.«S’ilît… de-mun»"
        cursor.execute("insert into %sbooze VALUES ('%s')" % (self.table_prefix, x))
        cursor.execute('select name from %sbooze' % self.table_prefix)
        self.assertEqual(x, cursor.fetchone()[0])
        con.close()

    def test_utf8(self):
        con = self._connect()
        try:
            cur = con.cursor()
            self.executeDDL1(cur)
            args = {'beer': '\xc4\xa5'}
            cur.execute('insert into %sbooze values (%%(beer)s)' % self.table_prefix, args)
            cur.execute('select name from %sbooze' % self.table_prefix)
            res = cur.fetchall()
            beer = res[0][0]
            self.assertEqual(beer, args['beer'], 'incorrect data retrieved')
        finally:
            con.close()

    def test_unicode(self):
        con = self._connect()
        try:
            cur = con.cursor()
            self.executeDDL1(cur)
            args = {'beer': '\N{latin small letter a with acute}'}
            encoded = args['beer']

            cur.execute('insert into %sbooze values (%%(beer)s)' % self.table_prefix, args)
            cur.execute('select name from %sbooze' % self.table_prefix)
            res = cur.fetchall()
            beer = res[0][0]
            self.assertEqual(beer, encoded, 'incorrect data retrieved')
        finally:
            con.close()

    def test_substring(self):
        con = self._connect()
        try:
            cur = con.cursor()
            self.executeDDL1(cur)
            args = {'beer': '"" \"\'\",\\"\\"\"\'\"'}
            cur.execute('insert into %sbooze values (%%(beer)s)' % self.table_prefix, args)
            cur.execute('select name from %sbooze' % self.table_prefix)
            res = cur.fetchall()
            beer = res[0][0]
            self.assertEqual(beer, args['beer'],
                             'incorrect data retrieved, got %s, should be %s' % (beer, args['beer']))
        finally:
            con.close()

    def test_escape(self):
        teststrings = [
            'abc\ndef',
            'abc\\ndef',
            'abc\\\ndef',
            'abc\\\\ndef',
            'abc\\\\\ndef',
            'abc"def',
            'abc""def',
            'abc\'def',
            'abc\'\'def',
            "abc\"def",
            "abc\"\"def",
            "abc'def",
            "abc''def",
            "abc\tdef",
            "abc\\tdef",
            "abc\\\tdef",
            "\\x"
        ]

        con = self._connect()
        cur = con.cursor()
        self.executeDDL1(cur)
        for i in teststrings:
            args = {'beer': i}
            cur.execute('insert into %sbooze values (%%(beer)s)' % self.table_prefix, args)
            cur.execute('select * from %sbooze' % self.table_prefix)
            row = cur.fetchone()
            cur.execute('delete from %sbooze where name=%%s' % self.table_prefix, i)
            self.assertEqual(i, row[0], 'newline not properly converted, got %s, should be %s' % (row[0], i))
        con.close()

    def test_non_ascii_string(self):
        con = self._connect()
        cur = con.cursor()
        self.executeDDL1(cur)
        input_ = '中文 zhōngwén'
        args = {'beer': input_}
        cur.execute('insert into %sbooze values (%%(beer)s)' % self.table_prefix,
                    args)
        cur.execute('select name from %sbooze' % self.table_prefix)
        res = cur.fetchall()
        returned = res[0][0]
        self.assertEqual(returned, input_)
        self.assertEqual(type(returned), str)
        con.close()

    def test_query_ending_with_comment(self):
        con = self._connect()
        cur = con.cursor()
        self.executeDDL1(cur)
        cur.execute("insert into %sbooze values ('foo')" % self.table_prefix)
        cur.execute('select * from %sbooze --This is a SQL comment' % self.table_prefix)
        # the above line should execute without problems
        self.assertEqual(1, cur.rowcount,
                         'queries ending in comments should be executed correctly')
        con.close()


if __name__ == '__main__':
    unittest.main()



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0.  If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
# Copyright 1997 - July 2008 CWI, August 2008 - 2016 MonetDB B.V.

"""
MonetDB Python API specific exceptions
"""

StandardError = Exception


class Warning(StandardError):
    """Exception raised for important warnings like data
    truncations while inserting, etc. It must be a subclass of
    the Python StandardError (defined in the module
    exceptions)."""
    pass


class Error(StandardError):
    """Exception that is the base class of all other error
    exceptions. You can use this to catch all errors with one
    single 'except' statement. Warnings are not considered
    errors and thus should not use this class as base. It must
    be a subclass of the Python StandardError (defined in the
    module exceptions)."""
    pass


class InterfaceError(Error):
    """Exception raised for errors that are related to the
    database interface rather than the database itself.  It
    must be a subclass of Error."""
    pass


class DatabaseError(Error):
    """Exception raised for errors that are related to the
    database.  It must be a subclass of Error."""
    pass


class DataError(DatabaseError):
    """Exception raised for errors that are due to problems with
    the processed data like division by zero, numeric value
    out of range, etc. It must be a subclass of DatabaseError."""
    pass


class OperationalError(DatabaseError):
    """Exception raised for errors that are related to the
    database's operation and not necessarily under the control
    of the programmer, e.g. an unexpected disconnect occurs,
    the data source name is not found, a transaction could not
    be processed, a memory allocation error occurred during
    processing, etc.  It must be a subclass of DatabaseError."""
    pass


class IntegrityError(DatabaseError):
    """Exception raised when the relational integrity of the
    database is affected, e.g. a foreign key check fails.  It
    must be a subclass of DatabaseError."""
    pass


class InternalError(DatabaseError):
    """Exception raised when the database encounters an internal
    error, e.g. the cursor is not valid anymore, the
    transaction is out of sync, etc.  It must be a subclass of
    DatabaseError."""
    pass


class ProgrammingError(DatabaseError):
    """Exception raised for programming errors, e.g. table not
    found or already exists, syntax error in the SQL
    statement, wrong number of parameters specified, etc.  It
    must be a subclass of DatabaseError."""
    pass


class NotSupportedError(DatabaseError):
    """Exception raised in case a method or database API was used which is not
    supported by the database, e.g. requesting a .rollback() on a connection
    that does not support transaction or has transactions turned off.  It must
    be a subclass of DatabaseError."""
    pass



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0.  If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
# Copyright 1997 - July 2008 CWI, August 2008 - 2016 MonetDB B.V.

import datetime
import unittest
import uuid
from pymonetdb.sql.monetize import convert
from pymonetdb.exceptions import ProgrammingError


class TestMonetize(unittest.TestCase):
    def test_str_subclass(self):
        class StrSubClass(str):
            pass
        x = StrSubClass('test')
        csub = convert(x)
        cstr = convert('test')
        self.assertEqual(csub, cstr)

    def test_unknown_type(self):
        class Unknown:
            pass
        x = Unknown()
        self.assertRaises(ProgrammingError, convert, x)

    def test_datetime(self):
        x = datetime.datetime(2017, 12, 6, 12, 30)
        self.assertEqual(convert(x), "TIMESTAMP '2017-12-06 12:30:00'")

    def test_datetime_tz(self):
        x = datetime.datetime(2017, 12, 6, 12, 30).replace(tzinfo=datetime.timezone(datetime.timedelta(hours=3)))
        self.assertEqual(convert(x), "TIMESTAMPTZ '2017-12-06 12:30:00+03:00'")

    def test_date(self):
        x = datetime.date(2017, 12, 6)
        self.assertEqual(convert(x), "DATE '2017-12-06'")

    def test_time(self):
        x = datetime.time(12, 5)
        self.assertEqual(convert(x), "TIME '12:05:00'")

    def test_time_tz(self):
        x = datetime.time(12, 5).replace(tzinfo=datetime.timezone(datetime.timedelta(hours=3)))
        self.assertEqual(convert(x), "TIMETZ '12:05:00+03:00'")

    def test_timedelta(self):
        x = datetime.timedelta(days=5, hours=2, minutes=10)
        self.assertEqual(convert(x), "INTERVAL '439800' SECOND")

    def test_uuids(self):
        x = uuid.UUID('334e6185-dd64-33d8-a052-d93371d3d20d')
        self.assertEqual(convert(x), "'334e6185-dd64-33d8-a052-d93371d3d20d'")



#!/usr/bin/env python

# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0.  If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
# Copyright 1997 - July 2008 CWI, August 2008 - 2016 MonetDB B.V.

import os
from setuptools import find_packages, setup

tests_require = ['pytest', 'mypy', 'pycodestyle', 'types-setuptools']

extras_require = {
    'test': tests_require,
    'doc': ['sphinx', 'sphinx_rtd_theme'],

}


def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()


__version__ = '1.6.1'

setup(
    name='pymonetdb',
    version=__version__,
    description='Native MonetDB client Python API',
    long_description=read('README.rst'),
    author='MonetDB BV',
    author_email='info@monetdb.org',
    url='http://www.monetdb.org/',
    packages=find_packages(),
    download_url='https://github.com/MonetDB/pymonetdb',
    classifiers=[
        "Topic :: Database",
        "Topic :: Database :: Database Engines/Servers",
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Mozilla Public License 2.0 (MPL 2.0)",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: Implementation :: PyPy",
    ],
    tests_require=tests_require,
    test_suite="tests",
    extras_require=extras_require,
)



import pickle
import tempfile
import re
import pdb
from typing import Any, TYPE_CHECKING


if TYPE_CHECKING:
    from pymonetdb.sql.cursors import Cursor


class LoopbackObject(object):
    def __init__(self, connection):
        self.__conn = connection

    def execute(self, query):
        self.__conn.execute("""
                  CREATE OR REPLACE FUNCTION export_parameters(*)
                  RETURNS TABLE(s STRING) LANGUAGE PYTHON
                  {
                      import inspect
                      import pickle
                      frame = inspect.currentframe();
                      args, _, _, values = inspect.getargvalues(frame);
                      dd = {x: values[x] for x in args};
                      del dd['_conn']
                      del dd['_columns']
                      del dd['_column_types']
                      return pickle.dumps(dd).hex();
                  };""")
        self.__conn.execute("""
                  SELECT *
                  FROM (%s) AS xx
                  LIMIT 1""" % query)
        query_description = self.__conn.description
        self.__conn.execute("""
                  SELECT *
                  FROM export_parameters ( (%s) );""" % query)
        rows = self.__conn.fetchall()
        arguments = pickle.loads(str(rows[0][0]))
        self.__conn.execute('DROP FUNCTION export_parameters;')
        if len(arguments) != len(query_description):
            raise Exception("Incorrect number of parameters!")
        result = dict()
        for j in range(len(arguments)):
            argname = "arg%d" % (j + 1)
            result[query_description[j][0]] = arguments[argname]
        return result


def debug(cursor: "Cursor", query: str, fname: str, sample: int = -1) -> Any:
    """ Locally debug a given Python UDF function in a SQL query
        using the PDB debugger. Optionally can run on only a
        sample of the input data, for faster data export.
    """

    # first gather information about the function
    cursor.execute("""
        SELECT func, type
        FROM functions
        WHERE language>=6 AND language <= 11 AND name='%s';""" % fname)
    data = cursor.fetchall()
    if len(data) == 0:
        raise Exception("Function not found!")

    # then gather the input arguments of the function
    cursor.execute("""
        SELECT args.name, args.type
        FROM args
        INNER JOIN functions ON args.func_id=functions.id
        WHERE functions.name='%s' AND args.inout=1
        ORDER BY args.number;""" % fname)
    input_types = cursor.fetchall()

    fcode = data[0][0]
    ftype = data[0][1]

    # now obtain the input columns
    arguments = exportparameters(cursor, ftype, fname, query, len(input_types), sample)

    arglist = "_columns, _column_types, _conn"
    cleaned_arguments = dict()
    for i in range(len(input_types)):
        argname = "arg%d" % (i + 1)
        if argname not in arguments:
            raise Exception("Argument %d not found!" % (i + 1))
        input_name = str(input_types[i][0])
        cleaned_arguments[input_name] = arguments[argname]
        arglist += ", %s" % input_name
    cleaned_arguments['_columns'] = arguments['_columns']
    cleaned_arguments['_column_types'] = arguments['_column_types']

    # create a temporary file for the function execution and run it
    with tempfile.NamedTemporaryFile() as f:
        fcode = fcode.strip()
        fcode = re.sub('^{', '', fcode)
        fcode = re.sub('};$', '', fcode)
        fcode = re.sub('^\n', '', fcode)
        function_definition = "def pyfun(%s):\n %s\n" % (
            arglist, fcode.replace("\n", "\n "))
        f.write(function_definition.encode('utf-8'))
        f.flush()
        compiled = compile(function_definition, f.name, 'exec')
        exec(compiled, globals(), locals())

        cleaned_arguments['_conn'] = LoopbackObject(cursor)
        pdb.set_trace()
        return locals()['pyfun'](*[], **cleaned_arguments)


def exportparameters(cursor, ftype, fname, query, quantity_parameters, sample):
    # type: (Cursor, str, str, str, Any, int) -> Any
    """ Exports the input parameters of a given UDF execution
        to the Python process. Used internally for .debug() and
        .export() functions.
    """

    # create a dummy function that only exports its parameters
    # using the pickle module
    if ftype == 5:
        # table producing function
        return_type = "TABLE(s STRING)"
    else:
        return_type = "STRING"

    if sample == -1:
        export_function = """
            CREATE OR REPLACE FUNCTION export_parameters(*)
            RETURNS %s LANGUAGE PYTHON
            {
                import inspect
                import pickle
                frame = inspect.currentframe();
                args, _, _, values = inspect.getargvalues(frame);
                dd = {x: values[x] for x in args};
                del dd['_conn']
                return pickle.dumps(dd).hex();
            };""" % return_type
    else:
        export_function = """
            CREATE OR REPLACE FUNCTION export_parameters(*)
            RETURNS %s LANGUAGE PYTHON
            {
            import inspect
            import pickle
            import numpy
            frame = inspect.currentframe();
            args, _, _, values = inspect.getargvalues(frame);
            dd = {x: values[x] for x in args};
            del dd['_conn']
            result = dict()
            argname = "arg1"
            x = numpy.arange(len(dd[argname]))
            x = numpy.random.choice(x,%s,replace=False)
            for i in range(len(dd)-2):
                argname = "arg" + str(i + 1)
                result = dd[argname]
                aux = []
                for j in range(len(x)):
                    aux.append(result[x[j]])
                dd[argname] = aux
                print(dd[argname])
            print(x)
            return pickle.dumps(dd).hex();
            };
            """ % (return_type, str(sample))

    if fname not in query:
        raise Exception("Function %s not found in query!" % fname)

    query = query.replace(fname, 'export_parameters')
    query = query.replace(';', ' sample 1;')

    cursor.execute(export_function)
    cursor.execute(query)
    input_data = cursor.fetchall()
    cursor.execute('DROP FUNCTION export_parameters;')
    if len(input_data) <= 0:
        raise Exception("Could not load input data!")

    bin_data = bytes.fromhex(input_data[0][0])
    arguments = pickle.loads(bin_data)

    if len(arguments) != quantity_parameters + 2:
        raise Exception("Incorrect amount of input arguments found!")

    return arguments


def export(cursor, query, fname, sample=-1, filespath='./'):
    """ Exports a Python UDF and its input parameters to a given
        file so it can be called locally in an IDE environment.
    """

    # first retrieve UDF information from the server
    cursor.execute("""
        SELECT func,type
        FROM functions
        WHERE language >= 6 AND language <= 11 AND name='%s';""" % fname)
    data = cursor.fetchall()
    cursor.execute("""
        SELECT args.name
        FROM args INNER JOIN functions ON args.func_id=functions.id
        WHERE functions.name='%s' AND args.inout=1
        ORDER BY args.number;""" % fname)
    input_names = cursor.fetchall()
    quantity_parameters = len(input_names)
    # fcode = data[0][0]
    ftype = data[0][1]
    parameter_list = []
    # exporting Python UDF Function
    if len(data) == 0:
        raise Exception("Function not found!")
    else:
        parameters = '('
        for x in range(0, len(input_names)):
            parameter = str(input_names[x]).split('\'')
            if x < len(input_names) - 1:
                parameter_list.append(parameter[1])
                parameters = parameters + parameter[1] + ','
            else:
                parameter_list.append(parameter[1])
                parameters = parameters + parameter[1] + '): \n'

        data = str(data[0]).replace('\\t', '\t').split('\\n')

        python_udf = 'import pickle \n \n \ndef ' + fname + parameters
        for x in range(1, len(data) - 1):
            python_udf = python_udf + '\t' + str(data[x]) + '\n'

    # exporting Columns as Binary Files
    arguments = exportparameters(cursor, ftype, fname, query, quantity_parameters, sample)
    result = dict()
    for i in range(len(arguments) - 2):
        argname = "arg%d" % (i + 1)
        result[parameter_list[i]] = arguments[argname]
    pickle.dump(result, open(filespath + 'input_data.bin', 'wb'))

    # loading Columns in Python & Call Function
    python_udf += '\n' + 'input_parameters = pickle.load(open(\''
    python_udf += filespath + 'input_data.bin\',\'rb\'))' + '\n'
    python_udf += fname + '('
    for i in range(0, quantity_parameters):
        if i < quantity_parameters - 1:
            python_udf += 'input_parameters[\''
            python_udf += parameter_list[i] + '\'],'
        else:
            python_udf += 'input_parameters[\''
            python_udf += parameter_list[i] + '\'])'

    file = open(filespath + fname + '.py', 'w')
    file.write(python_udf)
    file.close()



from unittest import TestCase
from tests.util import test_args
from pymonetdb.mapi import Connection


class TestMapi(TestCase):
    @classmethod
    def setUpClass(cls):
        cls.conn = Connection()
        cls.conn.connect(language='sql', **test_args)

    def test_set_size(self):
        query = 'sselect * from tables t1, tables t2;'
        for size in (2, 10):
            self.conn.set_reply_size(size)
            data = self.conn.cmd(query)
            cleaned = [i for i in data.split('\n') if i and not i[0] in '%&']
            self.assertEqual(len(cleaned), size)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0.  If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
# Copyright 1997 - July 2008 CWI, August 2008 - 2016 MonetDB B.V.

"""
definition of MonetDB column types, for more info:
http://www.monetdb.org/Documentation/Manuals/SQLreference/Datatypes
"""

CHAR = 'char'                      # (L) character string with length L
VARCHAR = 'varchar'                # (L) string with atmost length L
CLOB = 'clob'                      # string with no limit on size
BLOB = 'blob'                      # binary data

TINYINT = 'tinyint'                # 8 bit integer
SMALLINT = 'smallint'              # 16 bit integer
INT = 'int'                        # 32 bit integer
BIGINT = 'bigint'                  # 64 bit integer
HUGEINT = 'hugeint'                # 128 bit integer
SERIAL = 'serial'                  # special 64 bit integer sequence generator
REAL = 'real'                      # 32 bit floating point
DOUBLE = 'double'                  # 64 bit floating point
DECIMAL = 'decimal'                # (P,S)
BOOLEAN = 'boolean'                # boolean value

DATE = 'date'                      # (D) date
TIME = 'time'                      # (T) time of day
TIMESTAMP = 'timestamp'            # (T) date concatenated with unique time
MONTH_INTERVAL = 'month_interval'  # number of months
SEC_INTERVAL = 'sec_interval'      # number of seconds
DAY_INTERVAL = 'day_interval'      # number of seconds

URL = 'url'                        # url
INET = 'inet'                      # ipv4 address
UUID = 'uuid'                      # uuid
JSON = 'json'                      # JSON string
GEOMETRY = 'geometry'              # geometry string
GEOMETRYA = 'geometrya'            # geometry string
MBR = 'mbr'                        # mbr string
XML = 'xml'                        # xml

# Not on the website:
SHORTINT = 'shortint'
MEDIUMINT = 'mediumint'
LONGINT = 'longint'
OID = 'oid'
WRD = 'wrd'
FLOAT = 'float'                    # 64 bit floating point
TIMESTAMPTZ = 'timestamptz'        # with timezone
TIMETZ = 'timetz'                  # with timezone
INTERVAL = 'interval'              # (Q) a temporal interval

# full names and aliases, spaces are replaced with underscores
CHARACTER = CHAR
CHARACTER_VARYING = VARCHAR
CHARACHTER_LARGE_OBJECT = CLOB
BINARY_LARGE_OBJECT = BLOB
NUMERIC = DECIMAL
DOUBLE_PRECISION = DOUBLE



import os
from pymonetdb.exceptions import DatabaseError
from socket import gethostbyname
from unittest import TestCase
from pymonetdb import connect
from tests.util import test_args


class TestMapiUri(TestCase):

    def setUp(self):
        self.hostname = test_args['hostname']
        self.port = test_args['port']
        self.database = test_args['database']
        self.username = test_args['username']
        self.password = test_args['password']

    def attempt_connect(self, uri, username=None, password=None):
        args = dict(database=uri)
        if username:
            args['username'] = username
        if password:
            args['password'] = password
        connection = connect(autocommit=False, **args)
        cursor = connection.cursor()
        q = "select tag from sys.queue()"
        cursor.execute(q)
        cursor.fetchall()

    def test_no_uri(self):
        # test setUp and attempt_connect themselves
        self.attempt_connect(self.database, username=self.username, password=self.password)

    def test_full_mapi_uri(self):
        self.attempt_connect(f"mapi:monetdb://{self.hostname}:{self.port}/{self.database}",
                             username=self.username, password=self.password)

    def test_without_port(self):
        self.attempt_connect(f"mapi:monetdb://{self.hostname}/{self.database}",
                             username=self.username, password=self.password)

    def test_username_component(self):
        try:
            self.attempt_connect(f"mapi:monetdb://{self.hostname}:{self.port}/{self.database}",
                                 username="not" + self.username, password="not" + self.password)
        except DatabaseError:
            # expected to fail, username and password incorrect
            pass
        # override username and password parameters from within url
        s = f"mapi:monetdb://{self.username}:{self.password}@{self.hostname}:{self.port}/{self.database}"
        self.attempt_connect(s, username="not" + self.username, password="not" + self.password)

    def test_ipv4_address(self):
        try:
            # gethostbyname only resolves ipv4
            ip = gethostbyname(self.hostname)
        except Exception:
            # can't test, return success
            return
        self.attempt_connect(f"mapi:monetdb://{ip}:{self.port}/{self.database}",
                             username=self.username, password=self.password)

    def test_unix_domain_socket(self):
        sock_path = "/tmp/.s.monetdb.%i" % self.port
        if not os.path.exists(sock_path):
            return
        uri = f"mapi:monetdb://{sock_path}?database={self.database}"
        self.attempt_connect(uri, username=self.username, password=self.password)

    def test_unix_domain_socket_username(self):
        sock_path = "/tmp/.s.monetdb.%i" % self.port
        if not os.path.exists(sock_path):
            return
        uri = f"mapi:monetdb://{self.username}:{self.password}@{sock_path}?database={self.database}"
        self.attempt_connect(uri, username="not" + self.username, password="not" + self.password)



from unittest import TestCase, SkipTest
import tempfile
from typing import Optional
from tests.util import test_args
import pymonetdb
from pymonetdb.sql.connections import Connection


class TestUdf(TestCase):
    """
    To be able to run these tests python embedding needs to be enabled:

    monetdb create demo
    monetdb set embedpy=yes demo
    monetdb start demo

    or use embedpy3 for python3 support
    """
    conn: Optional[Connection] = None

    @classmethod
    def setUpClass(cls):
        cls.conn = pymonetdb.connect(**test_args)
        cls.cursor = cls.conn.cursor()

    @classmethod
    def tearDownClass(cls) -> None:
        if cls.conn:
            cls.conn.close()

    def test_debug_udf(self):
        self.cursor.execute("""
            CREATE FUNCTION test_python_udf(i INTEGER)
            RETURNS INTEGER
            LANGUAGE PYTHON {
                return i * 2;
            };
            """)
        # test if Python UDFs are enabled on the server
        try:
            self.cursor.execute('SELECT test_python_udf(1);')
        except pymonetdb.exceptions.OperationalError:
            raise SkipTest("We cant assume MonetDB is compiled with embbed python support enabled")

        else:
            result = self.cursor.fetchall()
            self.assertEqual(result, [(2,)])
            # test python debugging capabilities
            with tempfile.NamedTemporaryFile(delete=False) as f:
                fname = f.name
            self.cursor.export('SELECT test_python_udf(1)', 'test_python_udf', filespath=fname)
            fname += "test_python_udf.py"
            with open(fname) as f:
                code = f.read()
                self.assertEqual('test_python_udf(i)' in code, True)
